/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::SkillsMemory
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "SkillsMemory.h"

#include <ArmarXCore/core/ArmarXManager.h>
#include <ArmarXCore/core/ArmarXObjectScheduler.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/wm/memory_definitions.h>
#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>
#include <RobotAPI/libraries/armem_skills/aron_conversions.h>


namespace armarx
{

    SkillsMemory::SkillsMemory()
    {
    }


    armarx::PropertyDefinitionsPtr SkillsMemory::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs = new ComponentPropertyDefinitions(getConfigIdentifier());

        workingMemory().name() = "Skills";


        // Statechart Logging
        defs->optional(p.statechartCoreSegmentName, "StatechartCoreSegmentName", "Name of the core segment for statecharts.");
        defs->optional(p.statechartTransitionsProviderSegmentName, "TransitionsProviderSegmentName", "Name of the provider segment for statechart transitions.");
        defs->optional(p.statechartTransitionsTopicName, "tpc.sub.ProfilerListener", "Name of the ProfilerListenerInterface topics to subscribe.");

        return defs;
    }


    std::string SkillsMemory::getDefaultName() const
    {
        return "SkillsMemory";
    }


    void SkillsMemory::onInitComponent()
    {
        armarx::armem::server::wm::CoreSegment& c = workingMemory().addCoreSegment(
                    p.statechartCoreSegmentName,
                    armarx::armem::arondto::Statechart::Transition::toAronType());
        c.addProviderSegment("Transitions", armarx::armem::arondto::Statechart::Transition::toAronType());
    }


    void SkillsMemory::onConnectComponent()
    {
        statechartListener = createStatechartListener(p.statechartTransitionsTopicName);
    }


    void SkillsMemory::onDisconnectComponent()
    {
        if (statechartListener)
        {
            getArmarXManager()->removeObjectBlocking(statechartListener->getName());
        }
    }


    void SkillsMemory::onExitComponent()
    {
    }



    // WRITING
    armem::data::AddSegmentsResult SkillsMemory::addSegments(const armem::data::AddSegmentsInput& input, const Ice::Current&)
    {
        armem::data::AddSegmentsResult result = ReadWritePluginUser::addSegments({input}, p.core.addOnUsage);
        return result;
    }


    // READING
    // Inherited from Plugin


    IceInternal::Handle<StatechartListener>
    SkillsMemory::createStatechartListener(const std::string& topicName, const std::string& name)
    {
        const std::string name_ = name.empty() ? topicName + ".Listener" : name;
        ARMARX_DEBUG << "Registering StatechartListener '" << name_ << "' listening to topic '" << topicName << "'.";

        IceInternal::Handle<armarx::StatechartListener> listener = Component::create<armarx::StatechartListener>();
        listener->setName(name_);
        listener->setTopicName(topicName);

        // Callback for the transition listener
        listener->registerCallback(
            [this](const std::vector<StatechartListener::Transition>& transitions,
                   armarx::StatechartListener & source)
        {
            (void) source;  // Unused.
            this->reportTransitions(transitions);
        });

        try
        {
            getArmarXManager()->addObject(listener);
        }
        catch (const Ice::AlreadyRegisteredException& e)
        {
            ARMARX_ERROR << "The name '" << name_ << "' is already used. Please choose another one.\n"
                         << "Reason: " << e.what();
            getArmarXManager()->removeObjectBlocking(name_);
        }
        listener->getObjectScheduler()->waitForObjectState(armarx::ManagedIceObjectState::eManagedIceObjectStarted);

        return listener;
    }


    void
    SkillsMemory::reportTransitions(const std::vector<StatechartListener::Transition>& transitions)
    {
        for (const StatechartListener::Transition& t : transitions)
        {
            const std::string& entityName = getStatechartName(t.targetStateIdentifier);
            IceUtil::Time transitionTime = IceUtil::Time::microSeconds(t.timestamp);

            armem::EntityUpdate update;
            update.entityID = armem::MemoryID()
                              .withMemoryName(workingMemory().name())
                              .withCoreSegmentName(p.statechartCoreSegmentName)
                              .withProviderSegmentName(p.statechartTransitionsProviderSegmentName)
                              .withEntityName(entityName);

            update.timeCreated = transitionTime;
            armem::arondto::Statechart::Transition data;
            armem::toAron(data, t);
            update.instancesData.push_back(data.toAron());

            try
            {
                workingMemory().update(update);
            }
            catch (const armem::error::ArMemError& e)
            {
                ARMARX_WARNING << e.what();
            }
        }
    }


    std::string SkillsMemory::getStatechartName(std::string stateName)
    {
        const std::string delimiter = "->";
        const int maxLevels = 2;

        size_t pos;
        int levels = 0;
        std::string statechartName;
        while ((pos = stateName.find(delimiter)) != std::string::npos && levels < maxLevels)
        {
            if (levels != 0)
            {
                statechartName += delimiter;
            }
            statechartName += stateName.substr(0, pos);
            stateName.erase(0, pos + delimiter.length());
            levels++;
        }

        return statechartName;
    }
}
