// Header
#include "LongtermMemoryBase.h"

// ArmarX
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>


namespace armarx::armem::server::ltm
{
    void LongtermMemoryBase::setName(const std::string& name)
    {
        cache.name() = name;
        lut.name() = name;
    }

    armem::wm::Memory LongtermMemoryBase::getCacheAndLutNotConverted() const
    {
        std::lock_guard l(cache_mutex);
        std::lock_guard l2(lut_mutex);

        armem::wm::Memory m(lut.name());
        m.append(cache);
        m.append(lut);

        // debug output
        //lut.forEachSnapshot([](const auto & e)
        //{
        //    ARMARX_INFO << "The SNapshot: " << e.id().str() << " has size: " << e.size();
        //});

        return m;
    }

    void LongtermMemoryBase::append(const armem::wm::Memory& m)
    {
        TIMING_START(LTM_Append);
        ARMARX_INFO << "Append memory with name '" << m.name() << "' into the LTM with name '" << cache.name() << "'";

        std::lock_guard l(cache_mutex);
        cache.append(m);

        encodeAndStore();

        TIMING_END(LTM_Append);
    }

    void LongtermMemoryBase::checkUpdateLatestSnapshot(const armem::wm::EntitySnapshot& newSnapshot)
    {
        // update map of latestSnapshots
        if (auto it = latestSnapshots.find(newSnapshot.id().getEntityID().str()); it != latestSnapshots.end())
        {
            auto ptr = it->second;
            if (ptr->id().timestamp > newSnapshot.id().timestamp)
            {
                ptr = &newSnapshot;
            }
            // else ignore ==> no update
        }
        else
        {
            // no entry yet
            latestSnapshots.emplace(newSnapshot.id().getEntityID().str(), &newSnapshot);
        }
    }

    bool LongtermMemoryBase::containsCoreSegment(const MemoryID& coreSegmentID) const
    {
        //ARMARX_INFO << "Check if lut has core seg";
        if (lut.hasCoreSegment(coreSegmentID.coreSegmentName))
        {
            //ARMARX_INFO << "lus has core seg";
            return true;
        }
        return false;
    }

    bool LongtermMemoryBase::containsProviderSegment(const MemoryID& providerSegmentID) const
    {
        //ARMARX_INFO << "Check if lut has prov seg";
        if (lut.hasCoreSegment(providerSegmentID.coreSegmentName))
        {
            auto core = lut.getCoreSegment(providerSegmentID.coreSegmentName);
            if (core.hasProviderSegment(providerSegmentID.providerSegmentName))
            {
                //ARMARX_INFO << "lus has prov seg";
                return true;
            }
        }
        return false;
    }

    bool LongtermMemoryBase::containsEntity(const MemoryID& entityID) const
    {
        //ARMARX_INFO << "Check if lut has entity";
        if (lut.hasCoreSegment(entityID.coreSegmentName))
        {
            auto core = lut.getCoreSegment(entityID.coreSegmentName);
            if (core.hasProviderSegment(entityID.providerSegmentName))
            {
                auto prov = core.getProviderSegment(entityID.providerSegmentName);
                if (prov.hasEntity(entityID.entityName))
                {
                    //ARMARX_INFO << "lus has entity";
                    return true;
                }
            }
        }
        return false;
    }

    bool LongtermMemoryBase::containsSnapshot(const MemoryID& snapshotID) const
    {
        //ARMARX_INFO << "Check if lut has snapshot";
        if (lut.hasCoreSegment(snapshotID.coreSegmentName))
        {
            auto core = lut.getCoreSegment(snapshotID.coreSegmentName);
            if (core.hasProviderSegment(snapshotID.providerSegmentName))
            {
                auto prov = core.getProviderSegment(snapshotID.providerSegmentName);
                if (prov.hasEntity(snapshotID.entityName))
                {
                    auto entity = prov.getEntity(snapshotID.entityName);
                    if (entity.hasSnapshot(snapshotID.timestamp))
                    {
                        //ARMARX_INFO << "lut has snapshot";
                        return true;
                    }
                }
            }
        }
        return false;
    }
}
