// Header
#include "MemoryManager.h"

// STD / STL
#include <iostream>
#include <fstream>

// Simox
#include <SimoxUtility/json.h>

// ArmarX
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <RobotAPI/libraries/aron/core/navigator/data/container/Dict.h>
#include <RobotAPI/libraries/armem/core/wm/aron_conversions.h>
#include <RobotAPI/libraries/aron/converter/json/NLohmannJSONConverter.h>

namespace
{
    // check whether a string is a number (timestamp folders)
    bool isNumber(const std::string& s)
    {
        for (char const& ch : s)
        {
            if (std::isdigit(ch) == 0)
            {
                return false;
            }
        }
        return true;
    }
}

namespace armarx::armem::server::ltm::disk
{
    namespace fs = std::filesystem;

    bool MemoryManager::checkPath() const
    {
        // Check connection:
        ARMARX_INFO << "Checking Path";
        if (!fs::exists(basePathToMemory) || !fs::is_directory(basePathToMemory) || basePathToMemory.filename() != lut.name())
        {
            ARMARX_WARNING << deactivateSpam("PathIsNotValid")
                           << "The entered path is not valid. Please use a path leading to a memory folder with name: " << lut.name() << "."
                           << "\n\n";
            return false;
        }

        return true;
    }

    void MemoryManager::reload()
    {
        TIMING_START(LTM_Reload);
        ARMARX_INFO << "(Re)Loading a memory from: " << basePathToMemory.string();

        if (!checkPath())
        {
            // abort
            ARMARX_WARNING << "Could not (pre)load a memory from the filesystem.";
            return;
        }

        armem::wm::Memory temp(lut.id()); // a temporary client wm. We will append temp to the lut at the end of this metho (append ignores duplicate entries)
        ARMARX_INFO << "Loading memory: " << temp.id().str();

        // ///////////////////////////////
        // Iterate over core segments
        // ///////////////////////////////
        for (const auto& d : std::filesystem::directory_iterator(basePathToMemory))
            // Although this looks like code duplication, we need a distinguition between memories,
            // core, prov and entities because of a different structure
            // (only core and prov have same structure)
        {

            if (!d.is_directory())
            {
                continue;
            }

            std::string k = d.path().filename();
            if (temp.hasCoreSegment(k))
            {
                throw error::ArMemError("Somehow the (memory) container already contains the key k = " + k + ". This should not happen.");
            }

            // ///////////////////////////////
            // Add and iterate over core segments
            // ///////////////////////////////
            auto& cSeg = temp.addCoreSegment(k);
            for (const auto& d : std::filesystem::directory_iterator(d))
            {
                if (!d.is_directory())
                {
                    continue;
                }

                std::string k = d.path().filename();
                if (cSeg.hasProviderSegment(k))
                {
                    throw error::ArMemError("Somehow the (core) container already contains the key k = " + k + ". This should not happen.");
                }

                // ///////////////////////////////
                // Add and iterate over provider segments
                // ///////////////////////////////
                auto& pSeg = cSeg.addProviderSegment(k);
                for (const auto& d : std::filesystem::directory_iterator(d))
                {
                    if (!d.is_directory())
                    {
                        continue;
                    }

                    std::string k = d.path().filename();
                    if (pSeg.hasEntity(k))
                    {
                        throw error::ArMemError("Somehow the (provider) container already contains the key k = " + k + ". This should not happen.");
                    }

                    // ///////////////////////////////
                    // Add and iterate over entities
                    // ///////////////////////////////
                    auto& eSeg = pSeg.addEntity(k);
                    for (const auto& d : std::filesystem::directory_iterator(d))
                    {
                        if (!d.is_directory())
                        {
                            continue;
                        }

                        std::string k = d.path().filename();
                        if (!isNumber(k))
                        {
                            continue;
                        }

                        auto ts = armem::Time::microSeconds(std::stol(k));
                        // TODO catch exceptions?

                        if (eSeg.hasSnapshot(ts))
                        {
                            throw error::ArMemError("Somehow the (entity) container already contains the key k = " + k + ". This should not happen.");
                        }

                        // ///////////////////////////////
                        // Add and iterate over entities
                        // ///////////////////////////////
                        auto& sSeg = eSeg.addSnapshot(ts);
                        for (unsigned int i = 0; i < 10000; ++i) // ugly workaround to get the folders in the correct order
                        {
                            fs::path p = d / std::to_string(i);
                            if (!fs::exists(p) || !fs::is_directory(p))
                            {
                                // early stopping
                                break;
                            }

                            fs::path data = p / DATA_FILENAME;
                            if (!fs::exists(data) || !fs::is_regular_file(data))
                            {
                                // do not set data
                                continue;
                            }

                            // else we have an instance
                            /*std::ifstream ifs(data);
                            std::string file_content((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));

                            nlohmann::json j = nlohmann::json::parse(file_content);
                            auto aron = aron::converter::AronNlohmannJSONConverter::ConvertFromNlohmannJSON(j);*/

                            sSeg.addInstance();
                            //from_aron(aron, instance);
                        }
                    }
                }
            }
        }

        std::lock_guard l(cache_mutex); // we always take the cache mutex BEFORE the lut mutex! (otherwise we may have deadlocks)
        std::lock_guard l2(lut_mutex);
        lut.append(temp);
        ARMARX_INFO << "After reload memory " << lut.id().str() << " has size: " << lut.size();

        TIMING_END(LTM_Reload);
    }

    void MemoryManager::convert(armem::wm::Memory& m)
    {
        TIMING_START(LTM_Convert);
        if (!checkPath())
        {
            // abort
            ARMARX_WARNING << "Could not convert a memory from the filesystem.";
            return;
        }

        // update emtpy data ptr
        m.forEachCoreSegment([this](armem::wm::CoreSegment & e)
        {
            e.forEachProviderSegment([this](armem::wm::ProviderSegment & e)
            {
                e.forEachEntity([this](armem::wm::Entity & e)
                {
                    e.forEachSnapshot([this](armem::wm::EntitySnapshot & e)
                    {
                        // check whether data is nullptr
                        bool allDataIsNull = e.size() > 0;
                        e.forEachInstance([&allDataIsNull](armem::wm::EntityInstance & e)
                        {
                            if (e.data())
                            {
                                allDataIsNull = false;
                                return false; // means break
                            }
                            return true;
                        });

                        if (allDataIsNull) // an entry from the lut (probably... for now we assume that every entry either has data (cache) or has null (lut))
                        {
                            // Get data from mongodb
                            auto p = basePathToMemory / e.id().coreSegmentName / e.id().providerSegmentName / e.id().entityName / std::to_string(e.id().timestamp.toMicroSeconds());

                            if (fs::exists(p) && fs::is_directory(p))
                            {
                                for (unsigned int i = 0; i < e.size(); ++i)
                                {
                                    auto data = p / std::to_string(i) / DATA_FILENAME;

                                    if (fs::exists(data) && fs::is_regular_file(data))
                                    {
                                        auto& ins = e.getInstance(i);

                                        std::ifstream ifs(data);
                                        std::string file_content((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));
                                        nlohmann::json doc = nlohmann::json::parse(file_content);

                                        auto aron = aron::converter::AronNlohmannJSONConverter::ConvertFromNlohmannJSON(doc);

                                        wm::EntityInstance tmp(e.id().withInstanceIndex(i));
                                        from_aron(aron, tmp);

                                        ins.data() = tmp.data();
                                    }
                                }
                            }
                            // else leave snapshot untouched
                        }
                    });
                });
            });
        });
        TIMING_END(LTM_Convert);
    }

    void MemoryManager::encodeAndStore()
    {
        TIMING_START(LTM_Encode);
        if (!checkPath())
        {
            // abort
            ARMARX_WARNING << "Could not (pre)load a memory from the filesystem.";
            return;
        }

        std::lock_guard l(cache_mutex);

        // ///////////////////////////////
        // Iterate over core segments
        // ///////////////////////////////
        fs::path mPath = basePathToMemory;
        cache.forEachCoreSegment([&mPath](armem::wm::CoreSegment & e)
        {
            fs::path cPath = mPath / e.id().coreSegmentName;
            if (!fs::exists(cPath))
            {
                // not found
                fs::create_directory(cPath);
            }
            if (!fs::is_directory(cPath))
            {
                throw error::ArMemError("Could not create the (core) folder: " + cPath.string());
            }

            // ///////////////////////////////
            // Iterate over provider segments
            // ///////////////////////////////
            e.forEachProviderSegment([&cPath](armem::wm::ProviderSegment & e)
            {
                fs::path pPath = cPath / e.id().providerSegmentName;
                if (!fs::exists(pPath))
                {
                    // not found
                    fs::create_directory(pPath);
                }
                if (!fs::is_directory(pPath))
                {
                    throw error::ArMemError("Could not create the (provider) folder: " + pPath.string());
                }

                // ///////////////////////////////
                // Iterate over entities
                // ///////////////////////////////
                e.forEachEntity([&pPath](armem::wm::Entity & e)
                {
                    fs::path ePath = pPath / e.id().entityName;
                    if (!fs::exists(ePath))
                    {
                        // not found
                        fs::create_directory(ePath);
                    }
                    if (!fs::is_directory(ePath))
                    {
                        throw error::ArMemError("Could not create the (entity) folder: " + ePath.string());
                    }

                    // ///////////////////////////////
                    // Iterate over snapshots
                    // ///////////////////////////////
                    e.forEachSnapshot([&ePath](armem::wm::EntitySnapshot & e)
                    {
                        fs::path sPath = ePath / std::to_string(e.id().timestamp.toMicroSeconds());
                        if (!fs::exists(sPath))
                        {
                            // not found
                            fs::create_directory(sPath);
                        }
                        else
                        {
                            ARMARX_INFO << "The snapshot " << sPath.string() << " already exists. Ingoring it.";
                            return; // continue if already exists
                        }

                        if (!fs::is_directory(sPath))
                        {
                            throw error::ArMemError("Could not create the (timestamp) folder: " + sPath.string());
                        }

                        // ///////////////////////////////
                        // Iterate over instances
                        // ///////////////////////////////
                        e.forEachInstance([&sPath](armem::wm::EntityInstance & e)
                        {
                            fs::path iPath = sPath / std::to_string(e.id().instanceIndex);
                            if (!fs::exists(iPath))
                            {
                                // not found
                                fs::create_directory(iPath);
                            }
                            else
                            {
                                // This is strange, since we know, that the snapshot folder did not exist.
                                // However, we ignore and continue
                                ARMARX_INFO << "Somehow the instance folder " << iPath.string() << " already exists, although the snapshot folder was newly created. Ignore this.";
                                return;
                            }
                            if (!fs::is_directory(iPath))
                            {
                                throw error::ArMemError("Could not create the (instance) folder: " + iPath.string());
                            }

                            fs::path data = iPath / DATA_FILENAME;
                            if (fs::exists(data))
                            {
                                // Should not be the case. Anyway, if it happens, create new file!
                                fs::remove(data);
                            }

                            std::ofstream ofs;
                            ofs.open(data);

                            auto aron = std::make_shared<aron::datanavigator::DictNavigator>();
                            to_aron(aron, e);
                            nlohmann::json j = aron::converter::AronNlohmannJSONConverter::ConvertToNlohmannJSON(aron);

                            ofs << j.dump(2);
                            ofs.close();
                        });
                    });
                });
            });

        });

        // what to do with clear text data after encoding?
        // TODO!

        // Finaly clear cache and put reference to lut
        cache.forEachInstance([](armem::wm::EntityInstance & i)
        {
            i.data() = nullptr;
        });

        std::lock_guard l2(lut_mutex);
        lut.append(cache);
        cache.clear();

        TIMING_END(LTM_Encode);
    }
}
