// Header
#include "MemoryManager.h"

// Simox
#include <SimoxUtility/json.h>

// ArmarX
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/aron/core/navigator/data/container/Dict.h>
#include <RobotAPI/libraries/armem/core/wm/aron_conversions.h>
#include <RobotAPI/libraries/aron/converter/json/NLohmannJSONConverter.h>


namespace armarx::armem::server::ltm::mongodb
{
    namespace bsoncxxbuilder = bsoncxx::builder::stream;
    namespace bsoncxxdoc = bsoncxx::document;

    PoolClientPtr MemoryManager::checkConnection() const
    {
        // Check connection:
        ARMARX_INFO << "Checking connection";
        if (!ConnectionManager::ConnectionIsValid(dbsettings))
        {
            ARMARX_WARNING << deactivateSpam("ConnectionIsNotValid")
                           << "The connection to mongocxx for ltm '" << cache.name() << "' is not valid. Settings are: " << dbsettings.toString()
                           << "\nTo start it, run e.g.: \n"
                           << "armarx memory start"
                           << "\n\n";
            return nullptr;
        }

        auto& pool = ConnectionManager::Connect(dbsettings);
        auto client = pool.acquire();

        return client;
    }

    void MemoryManager::reload()
    {
        TIMING_START(LTM_Reload);
        ARMARX_INFO << "(Re)Establishing connection to: " << dbsettings.toString();

        auto client = checkConnection();
        if (!client)
        {
            // abort
            ARMARX_WARNING << "A connection to: " << dbsettings.toString() << " is not possible. Could not (pre)load data from mongodb.";
            return;
        }

        auto databases = client->list_databases();
        for (const auto& doc : databases)
        {
            auto el = doc["name"];
            ARMARX_INFO << "Found Memory-Collection in MongoDB: " << el.get_utf8().value;
        }

        armem::wm::Memory temp(lut.id()); // a temporary client wm. We will append temp to the lut at the end of this metho (append ignores duplicate entries)
        mongocxx::database db = client->database(dbsettings.database);

        ARMARX_INFO << "Loading memory: " << temp.id().str();

        // ///////////////////////////////
        // Iterate over core segments
        // ///////////////////////////////
        mongocxx::collection mColl = db.collection(temp.id().str());
        mongocxx::cursor cursor = mColl.find({});
        for (const auto& doc : cursor)  // Although this looks like code duplication, we need a distinguition between memories,
            // core, prov and entities because of a different structure
            // (only core and prov have same structure)
        {
            auto el = doc[MONGO_FOREIGN_KEY];
            auto foreignKey = el.get_utf8().value;

            MemoryID i((std::string) foreignKey);
            if (i.memoryName != temp.id().memoryName)
            {
                throw error::InvalidMemoryID(i, "A MemoryID in mongodb was invalid. Found the wrong memory name: " + i.str());
            }

            std::string k = i.coreSegmentName;

            if (temp.hasCoreSegment(k))
            {
                throw error::ArMemError("Somehow the (memory) container already contains the key k = " + k + ". Do you have double entries in mongodb?");
            }
            else
            {
                // ///////////////////////////////
                // Add and iterate over provider segments
                // ///////////////////////////////
                auto& cSeg = temp.addCoreSegment(k);
                mongocxx::collection cColl = db.collection(cSeg.id().str());
                mongocxx::cursor cursor = cColl.find({});
                for (const auto& doc : cursor)
                {
                    auto el = doc[MONGO_FOREIGN_KEY];
                    auto foreignKey = el.get_utf8().value;

                    MemoryID i((std::string) foreignKey);
                    if (i.coreSegmentName != cSeg.id().coreSegmentName || i.memoryName != cSeg.id().memoryName)
                    {
                        throw error::InvalidMemoryID(i, "A MemoryID in mongodb was invalid. Found the wrong memory name: " + i.str());
                    }

                    std::string k = i.providerSegmentName;
                    if (cSeg.hasProviderSegment(k))
                    {
                        throw error::ArMemError("Somehow the (core segment) container already contains the key k = " + k + ". Do you have double entries in mongodb?");
                    }
                    else
                    {
                        // ///////////////////////////////
                        // Add and iterate over entities
                        // ///////////////////////////////
                        auto& pSeg = cSeg.addProviderSegment(k);
                        mongocxx::collection pColl = db.collection(pSeg.id().str());
                        mongocxx::cursor cursor = pColl.find({});
                        for (const auto& doc : cursor)
                        {
                            auto el = doc[MONGO_FOREIGN_KEY];
                            auto foreignKey = el.get_utf8().value;

                            MemoryID i((std::string) foreignKey);
                            if (i.providerSegmentName != pSeg.id().providerSegmentName || i.coreSegmentName != pSeg.id().coreSegmentName || i.memoryName != pSeg.id().memoryName)
                            {
                                throw error::InvalidMemoryID(i, "A MemoryID in mongodb was invalid. Found the wrong memory name: " + i.str());
                            }

                            std::string k = i.entityName;
                            if (pSeg.hasEntity(k))
                            {
                                throw error::ArMemError("Somehow the (provider segment) container already contains the key k = " + k + ". Do you have double entries in mongodb?");
                            }
                            else
                            {
                                // ///////////////////////////////
                                // Add and iterate over snapshots
                                // ///////////////////////////////
                                auto& eSeg = pSeg.addEntity(k);
                                mongocxx::collection eColl = db.collection(eSeg.id().str());
                                mongocxx::cursor cursor = eColl.find({});
                                for (const auto& doc : cursor)
                                {
                                    auto ts = armem::Time::microSeconds(doc[MONGO_TIMESTAMP].get_int64().value);
                                    auto& newSnapshot = eSeg.addSnapshot(ts);

                                    auto i = doc[MONGO_INSTANCES];
                                    unsigned long length = std::distance(i.get_array().value.begin(), i.get_array().value.end());
                                    ARMARX_INFO << "The array legth for an snapshot '" << newSnapshot.id().str() << "' is: " << length;

                                    for (unsigned long i = 0; i < length; ++i)
                                    {
                                        // add an empty instance as reference
                                        newSnapshot.addInstance({});
                                    }


                                    // check to update lastSnapshot map
                                    checkUpdateLatestSnapshot(newSnapshot);
                                }
                            }
                        }
                    }
                }
            }
        }

        std::lock_guard l(cache_mutex); // we always take the cache mutex BEFORE the lut mutex! (otherwise we may have deadlocks)
        std::lock_guard l2(lut_mutex);
        lut.append(temp);
        ARMARX_INFO << "After reload memory " << lut.id().str() << " has size: " << lut.size();

        TIMING_END(LTM_Reload);
    }

    void MemoryManager::convert(armem::wm::Memory& m)
    {
        TIMING_START(LTM_Convert);

        auto client = checkConnection();
        if (!client)
        {
            // abort
            ARMARX_WARNING << "A connection to: " << dbsettings.toString() << " is not possible. Could not convert information, therefore returning leaving input untouched.";
            return;
        }
        mongocxx::database db = client->database(dbsettings.database);

        // update emtpy data ptr
        m.forEachCoreSegment([&db](armem::wm::CoreSegment & e)
        {
            e.forEachProviderSegment([&db](armem::wm::ProviderSegment & e)
            {
                e.forEachEntity([&db](armem::wm::Entity & e)
                {
                    e.forEachSnapshot([&db](armem::wm::EntitySnapshot & e)
                    {
                        // check whether data is nullptr
                        bool allDataIsNull = e.size() > 0;
                        e.forEachInstance([&allDataIsNull](armem::wm::EntityInstance & e)
                        {
                            if (e.data())
                            {
                                allDataIsNull = false;
                                return false; // means break
                            }
                            return true;
                        });

                        if (allDataIsNull) // an entry from the lut (probably... for now we assume that every entry either has data (cache) or has null (lut))
                        {
                            // Get data from mongodb
                            auto eColl = db.collection(e.id().getEntityID().str());
                            auto q = bsoncxxbuilder::document{} << std::string(MONGO_TIMESTAMP) << e.id().timestamp.toMicroSeconds() << bsoncxxbuilder::finalize;
                            auto res = eColl.find_one(q.view());

                            if (!res)
                            {
                                throw error::ArMemError("Could not load an instance from the memory '" + e.id().getEntityID().str() + "'. Tried to access: " + e.id().getEntitySnapshotID().str());
                            }

                            // convert full json of this entry
                            nlohmann::json json = nlohmann::json::parse(bsoncxx::to_json(*res));
                            nlohmann::json instances = json[MONGO_INSTANCES];

                            if (instances.size() != e.size())
                            {
                                throw error::ArMemError("The size of the mongodb entity entry at id " + e.id().getEntitySnapshotID().str() + " has wrong size. Expected: " + std::to_string(e.size()) + " but got: " + std::to_string(instances.size()));
                            }

                            for (unsigned int i = 0; i < e.size(); ++i)
                            {
                                auto& ins = e.getInstance(i);

                                // get ionstance json
                                nlohmann::json doc = instances[i];
                                auto aron = aron::converter::AronNlohmannJSONConverter::ConvertFromNlohmannJSON(doc);

                                // remove metadata
                                wm::EntityInstance tmp(e.id().withInstanceIndex(i));
                                from_aron(aron, tmp);

                                // set data
                                ins.data() = tmp.data();
                            }
                        }
                    });
                });
            });
        });
        TIMING_END(LTM_Convert);
    }

    void MemoryManager::encodeAndStore()
    {
        TIMING_START(LTM_Encode);
        ARMARX_INFO << "Encode cache " << cache.id().str() << " with size: " << cache.size();

        auto client = checkConnection();
        if (!client)
        {
            // abort
            ARMARX_WARNING << "A connection to: " << dbsettings.toString() << " is not possible. Could not consolidate data from cache to mongodb.";
            return;
        }

        std::lock_guard l(cache_mutex);
        mongocxx::database db = client->database(dbsettings.database);
        auto mColl = db.collection(cache.id().str());

        // ///////////////////////////////
        // Iterate over core segments
        // ///////////////////////////////
        cache.forEachCoreSegment([this, &db, &mColl](armem::wm::CoreSegment & e)
        {
            auto cColl = db.collection(e.id().str());

            if (!containsCoreSegment(mColl, e.id()))
            {
                // not found
                mongodbInsertForeignKey(mColl, e.id().str());
            }

            // ///////////////////////////////
            // Iterate over provider segments
            // ///////////////////////////////
            e.forEachProviderSegment([this, &db, &cColl](armem::wm::ProviderSegment & e)
            {
                auto pColl = db.collection(e.id().str());

                if (!containsProviderSegment(cColl, e.id()))
                {
                    // not found
                    mongodbInsertForeignKey(cColl, e.id().str());
                }

                // ///////////////////////////////
                // Iterate over each segments
                // ///////////////////////////////
                e.forEachEntity([this, &db, &pColl](armem::wm::Entity & e)
                {
                    auto eColl = db.collection(e.id().str());

                    if (!containsEntity(pColl, e.id()))
                    {
                        // not found
                        mongodbInsertForeignKey(pColl, e.id().str());
                    }

                    // ///////////////////////////////
                    // Iterate over snapshots
                    // ///////////////////////////////
                    e.forEachSnapshot([this, &eColl](armem::wm::EntitySnapshot & e)
                    {
                        if (!this->containsSnapshot(eColl, e.id()))
                        {
                            // timestamp not found in mongodb ==> new entry
                            bsoncxxbuilder::document builder{};
                            auto in_array = builder
                                            << std::string(MONGO_ID) << e.id().str()
                                            << std::string(MONGO_TIMESTAMP) << e.id().timestamp.toMicroSeconds()
                                            << std::string(MONGO_INSTANCES);
                            auto array_builder = bsoncxx::builder::basic::array{};

                            e.forEachInstance([&array_builder](const wm::EntityInstance & e)
                            {
                                auto aron = std::make_shared<aron::datanavigator::DictNavigator>();
                                to_aron(aron, e);
                                nlohmann::json j = aron::converter::AronNlohmannJSONConverter::ConvertToNlohmannJSON(aron);

                                auto doc_value = bsoncxx::from_json(j.dump(2));
                                array_builder.append(doc_value);
                            });

                            auto after_array = in_array << array_builder;
                            bsoncxx::document::value doc = after_array << bsoncxx::builder::stream::finalize;
                            ARMARX_INFO << "Insert into mongodb: " << e.id().timestamp.toMicroSeconds();
                            eColl.insert_one(doc.view());

                            // check to update lastSnapshot map
                            checkUpdateLatestSnapshot(e);
                        }
                        else
                        {
                            ARMARX_INFO << "Timestamp already found for id: " << e.id().str();
                            // timestamp already in mongodb. Ignore it
                        }
                    });
                });
            });
        });

        // what to do with clear text data after encoding?
        // TODO!

        // Finaly clear cache and put reference to lut
        cache.forEachInstance([](armem::wm::EntityInstance & i)
        {
            i.data() = nullptr;
        });

        std::lock_guard l2(lut_mutex);
        lut.append(cache);
        cache.clear();

        TIMING_END(LTM_Encode);
    }

    void MemoryManager::mongodbInsertForeignKey(mongocxx::collection& coll, const std::string& key)
    {
        auto q = bsoncxxbuilder::document{} << std::string(MONGO_FOREIGN_KEY) << key << bsoncxxbuilder::finalize;
        coll.insert_one(q.view());
    }

    bool MemoryManager::mongodbContainsForeignKey(mongocxx::collection& coll, const std::string& key) const
    {
        // check mongodb
        auto q = bsoncxxbuilder::document{} << std::string(MONGO_FOREIGN_KEY) << key << bsoncxxbuilder::finalize;
        auto res = coll.find_one(q.view());
        return (bool) res;
    }

    bool MemoryManager::containsCoreSegment(mongocxx::collection& mColl, const MemoryID& coreSegmentID) const
    {
        if (Base::containsCoreSegment(coreSegmentID))
        {
            return true;
        }
        return mongodbContainsForeignKey(mColl, coreSegmentID.str());
    }

    bool MemoryManager::containsProviderSegment(mongocxx::collection& cColl, const MemoryID& providerSegmentID) const
    {
        if (Base::containsProviderSegment(providerSegmentID))
        {
            return true;
        }
        return mongodbContainsForeignKey(cColl, providerSegmentID.str());
    }

    bool MemoryManager::containsEntity(mongocxx::collection& pColl, const MemoryID& entityID) const
    {
        if (Base::containsEntity(entityID))
        {
            return true;
        }
        return mongodbContainsForeignKey(pColl, entityID.str());
    }

    bool MemoryManager::containsSnapshot(mongocxx::collection& eColl, const MemoryID& snapshotID) const
    {
        if (Base::containsSnapshot(snapshotID))
        {
            return true;
        }

        // check mongodb
        auto q = bsoncxxbuilder::document{} << std::string(MONGO_TIMESTAMP) << snapshotID.timestamp.toMicroSeconds() << bsoncxxbuilder::finalize;
        auto res = eColl.find_one(q.view());
        return (bool) res;
    }
}
