/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Rainer Kartmann (rainer dot kartmann at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "Image.h"

#include <SimoxUtility/meta/type_name.h>


namespace armarx::aron::cppserializer::serializer
{

    const std::map<typenavigator::ImagePixelType, std::string>
    ImageSerializer::pixelTypeMap =
    {
        { typenavigator::ImagePixelType::Rgb24,   "CV_8UC3" },
        { typenavigator::ImagePixelType::Depth32, "CV_32FC1"},
    };


    // constructors
    ImageSerializer::ImageSerializer(const typenavigator::ImageNavigatorPtr& n) :
        detail::NDArraySerializerBase<typenavigator::ImageNavigator, ImageSerializer>(
            "cv::Mat",
            simox::meta::get_type_name<data::AronNDArray>(),
            simox::meta::get_type_name<type::AronImage>(),
            n)
    {
        ARMARX_CHECK_NOT_NULL(typenavigator);
    }


    CppBlockPtr ImageSerializer::getResetHardBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = std::make_shared<CppBlock>();

        std::string typeConstant;
        {
            if (auto it = pixelTypeMap.find(typenavigator->getPixelType()); it != pixelTypeMap.end())
            {
                typeConstant = it->second;
            }
            else
            {
                std::stringstream ss;
                ss << "Pixel type '" << static_cast<int>(typenavigator->getPixelType()) << "' "
                   << "is not mapped to an OpenCV pixel type.";
                throw error::AronException("ImageSerializer", __FUNCTION__, ss.str());
            }
        }

        block_if_data->addLine(accessor + ".create(0, 0, " + typeConstant + ");");
        return this->ResolveMaybeResetHardBlock(accessor, block_if_data, this->typenavigator);
    }


    CppBlockPtr ImageSerializer::getResetSoftBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = std::make_shared<CppBlock>();

        block_if_data->addLine(accessor + " = 0;");

        return ResolveMaybeResetSoftBlock(accessor, block_if_data, typenavigator);
    }


    CppBlockPtr ImageSerializer::getWriteTypeBlock(const std::string& accessor) const
    {
        CppBlockPtr b = std::make_shared<CppBlock>();

        std::string typeName;
        switch (typenavigator->getPixelType())
        {
#define CASE( typeConstant ) \
case typeConstant : \
    typeName = "::armarx::aron::typenavigator::ImageNavigator::pixelTypeToName( " #typeConstant " )"; \
    break;

                CASE(::armarx::aron::typenavigator::ImagePixelType::Rgb24)
                CASE(::armarx::aron::typenavigator::ImagePixelType::Depth32)

#undef CASE
        }

        b->addLine("w.writeImage({" + typeName + ", " + MAYBE_TO_STRING(typenavigator->getMaybe()) + "});  // of " + accessor);
        return b;
    }


    CppBlockPtr ImageSerializer::getWriteBlock(const std::string& accessor) const
    {
        CppBlockPtr block = std::make_shared<CppBlock>();

        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string accessor_shape = "shape";

        block->addLine("std::vector<int> " + accessor_shape + "(" + accessor + ".size.p, " + accessor + ".size.p + " + accessor + ".dims);");
        block->addLine(accessor_shape + ".push_back(" + accessor + nextEl() + "elemSize());");


#if 0  // Target code:
        std::string type;
        switch (the_depth32_image.type())
        {
            case CV_8UC3:
                type = ::armarx::aron::typenavigator::ImageNavigator::pixelTypeToName(::armarx::aron::typenavigator::ImagePixelType::RGB24);
                break;
            case CV_32FC1:
                type = ::armarx::aron::typenavigator::ImageNavigator::pixelTypeToName(::armarx::aron::typenavigator::ImagePixelType::DEPTH32);
                break;
            default:
            {
                std::stringstream ss;
                ss << "OpenCV image type " << the_depth32_image.type() << " cannot be serialized.";
                throw ::armarx::aron::error::AronException("ImageTest", __FUNCTION__, ss.str());
            }
        }
#endif

        block->addLine("std::string arrayType;");
        block->addLine("switch (" + accessor + ".type())");
        {
            CppBlockPtr block_switch = std::make_shared<CppBlock>();

#define CASE( typeConstant ) \
    block_switch->addLine("case " + pixelTypeMap.at(typeConstant) + ":"); \
    block_switch->addLine("arrayType = ::armarx::aron::typenavigator::ImageNavigator::pixelTypeToName(" #typeConstant ");"); \
    block_switch->addLine("break;")

            CASE(::armarx::aron::typenavigator::ImagePixelType::Rgb24);
            CASE(::armarx::aron::typenavigator::ImagePixelType::Depth32);

#undef CASE

            block_switch->addLine("default:");
            {
                CppBlockPtr block_default = std::make_shared<CppBlock>();
                block_default->addLine("std::stringstream ss;");
                block_default->addLine("ss << \"OpenCV image type \" << " + accessor + ".type() << \" cannot be serialized.\";");
                block_default->addLine("throw ::armarx::aron::error::AronException(\"" + typenavigator->getName() + "\", __FUNCTION__, ss.str());");

                block_switch->addBlock(block_default);
            }
            block->addBlock(block_switch);
        }


        block->addLine("w.writeNDArray(" + accessor_shape + ", arrayType, reinterpret_cast<const unsigned char*>(" + accessor + nextEl() + "data)); // of " + accessor);

        return ResolveMaybeWriteBlock(accessor, block, typenavigator);
    }


    CppBlockPtr ImageSerializer::getReadBlock(const std::string& accessor) const
    {
        const std::string escaped_accessor = EscapeAccessor(accessor);
        const std::string read_start_result_accessor = escaped_accessor + READ_START_RETURN_TYPE_ACCESSOR;

#if 0   // Target code
        auto the_rgb24_image_return_type = r.readStartNDArray(); // of the_rgb24_image

        if (!the_rgb24_image_return_type.dims.empty())
        {
            std::vector<int> shape{the_rgb24_image_return_type.dims.begin(), std::prev(the_rgb24_image_return_type.dims.end())};

            ::armarx::aron::typenavigator::ImagePixelType pixelType = ::armarx::aron::typenavigator::ImageNavigator::pixelTypeFromName(the_rgb24_image_return_type.type);
            int cvMatType = 0;
            switch (pixelType)
            {
                case ::armarx::aron::typenavigator::ImagePixelType::RGB24:
                    cvMatType =  CV_8UC3;
                case ::armarx::aron::typenavigator::ImagePixelType::DEPTH32:
                    cvMatType =  CV_32FC1;
                default:
                {
                    std::stringstream ss;
                    ss << "NDArray Type '" << the_rgb24_image_return_type.type << "' cannot be deserialized as image.";
                    throw ::armarx::aron::error::AronException("AronImageType", __FUNCTION__, ss.str());
                }
            }

            the_rgb24_image.create(shape, cvMatType);
            r.readEndNDArray(reinterpret_cast<unsigned char*>(the_rgb24_image.data)); // of the_rgb24_image
        }
#endif

        CppBlockPtr block = std::make_shared<CppBlock>();

        block->addLine("if (!" + read_start_result_accessor + ".dims.empty())");
        {
            CppBlockPtr block_if = std::make_shared<CppBlock>();

            block_if->addLine("std::vector<int> shape{" + read_start_result_accessor + ".dims.begin(), std::prev(" + read_start_result_accessor + ".dims.end())};");

            block_if->addLine("::armarx::aron::typenavigator::ImagePixelType pixelType = ::armarx::aron::typenavigator::ImageNavigator::pixelTypeFromName(" + read_start_result_accessor + ".type);");
            block_if->addLine("int cvMatType = 0;");
            block_if->addLine("switch (pixelType)");
            {
                CppBlockPtr block_switch = std::make_shared<CppBlock>();

#define CASE( typeConstant ) \
    block_switch->addLine("case " #typeConstant ":"); \
    block_switch->addLine("cvMatType = " + pixelTypeMap.at(typeConstant) + ";"); \
    block_switch->addLine("break;")

                CASE(::armarx::aron::typenavigator::ImagePixelType::Rgb24);
                CASE(::armarx::aron::typenavigator::ImagePixelType::Depth32);

#undef CASE

                block_switch->addLine("default:");
                {
                    CppBlockPtr block_default = std::make_shared<CppBlock>();
                    block_default->addLine("std::stringstream ss;");
                    block_default->addLine("ss << \"NdArray Type '\" << " + read_start_result_accessor + ".type << \"' cannot be deserialized as image.\";");
                    block_default->addLine("throw ::armarx::aron::error::AronException(\"AronImageType\", __FUNCTION__, ss.str());");

                    block_switch->addBlock(block_default);
                }
                block_if->addBlock(block_switch);
            }

            block_if->addLine(accessor + ".create(shape, cvMatType);");
            block_if->addLine("r.readEndNDArray(reinterpret_cast<unsigned char*>(" + accessor + nextEl() + "data)); // of " + accessor);
            block->addBlock(block_if);
        }

        return ResolveMaybeReadBlock(accessor, "r.readStartNDArray()", block, typenavigator, true);
    }


    CppBlockPtr ImageSerializer::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr block_if_data = std::make_shared<CppBlock>();
        block_if_data->addLine("if (cv::countNonZero(" + accessor + " != " + otherInstanceAccessor + ") != 0)");
        block_if_data->addLineAsBlock("return false;");
        return ResolveMaybeEqualsBlock(accessor, otherInstanceAccessor, block_if_data, typenavigator);
    }
}


