#include "TypedDataVisitor.h"

#include <SimoxUtility/meta/type_name.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>


namespace armarx::aron::visitor
{


    TypedDataVisitor::TypedDataVisitor()
    {
    }

    TypedDataVisitor::~TypedDataVisitor()
    {
    }

    bool TypedDataVisitor::applyTo(TypeNavigator& type, DataNavigator& data)
    {
        return this->applyTo("", type, data);
    }

    template <class DataT, class TypeT>
    bool TypedDataVisitor::applyToCollection(TypeT& type, const std::string& key, DataNavigator& data)
    {
        DataT& dataCast = dynamic_cast<DataT&>(data);
        if (!visitEnter(type, key, dataCast))
        {
            return false;
        }
        if (!applyToChildren(type, dataCast))
        {
            return false;
        }
        if (!visitExit(type, key, dataCast))
        {
            return false;
        }
        return true;
    }

    template <class ToT, class FromT>
    ToT cast(const FromT& from)
    {
        ToT to;
        to.setValue(from.getValue());
        return to;
    }
    template <class FromDataT>
    std::optional<datanavigator::LongNavigator> castToLong(const FromDataT& data)
    {
        if (auto d = dynamic_cast<const datanavigator::IntNavigator*>(&data))
        {
            return cast<datanavigator::LongNavigator>(*d);
        }
        return std::nullopt;
    }
    template <class FromDataT>
    std::optional<datanavigator::IntNavigator> castToInt(const FromDataT& data)
    {
        if (auto d = dynamic_cast<const datanavigator::LongNavigator*>(&data))
        {
            return cast<datanavigator::IntNavigator>(*d);
        }
        return std::nullopt;
    }


    bool TypedDataVisitor::applyTo(const std::string& key, TypeNavigator& type, DataNavigator& data)
    {
        if (auto t = dynamic_cast<ObjectTypeNavigator*>(&type))
        {
            return applyToCollection<DictDataNavigator>(*t, key, data);
        }
        else if (auto t = dynamic_cast<DictTypeNavigator*>(&type))
        {
            return applyToCollection<DictDataNavigator>(*t, key, data);
        }
        else if (auto t = dynamic_cast<ListTypeNavigator*>(&type))
        {
            return applyToCollection<ListDataNavigator>(*t, key, data);
        }
        else if (auto t = dynamic_cast<TupleTypeNavigator*>(&type))
        {
            return applyToCollection<ListDataNavigator>(*t, key, data);
        }

        ARMARX_CHECK_EQUAL(type.childrenSize(), 0) << simox::meta::get_type_name(type);
        ARMARX_CHECK_EQUAL(data.childrenSize(), 0) << simox::meta::get_type_name(data);

        try
        {

            if (auto t = dynamic_cast<BoolTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<BoolDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<DoubleTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<DoubleDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<FloatTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<FloatDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<IntTypeNavigator*>(&type))
            {
                if (auto cast = castToInt(data))
                {
                    return visit(*t, key, cast.value());
                }
                else
                {
                    return visit(*t, key, dynamic_cast<IntDataNavigator&>(data));
                }
            }
            else if (auto t = dynamic_cast<LongTypeNavigator*>(&type))
            {
                if (auto cast = castToLong(data))
                {
                    return visit(*t, key, cast.value());
                }
                else
                {
                    return visit(*t, key, dynamic_cast<LongDataNavigator&>(data));
                }
            }
            else if (auto t = dynamic_cast<StringTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<StringDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<TimeTypeNavigator*>(&type))
            {
                if (auto cast = castToLong(data))
                {
                    return visit(*t, key, cast.value());
                }
                else
                {
                    return visit(*t, key, dynamic_cast<LongDataNavigator&>(data));
                }
            }

            if (auto t = dynamic_cast<PCLPointCloudTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<NDArrayDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<EigenMatrixTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<NDArrayDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<EigenQuaternionTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<NDArrayDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<ImageNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<NDArrayDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<IVTCByteImageTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<NDArrayDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<OpenCVMatTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<NDArrayDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<PoseTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<NDArrayDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<PositionTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<NDArrayDataNavigator&>(data));
            }
            else if (auto t = dynamic_cast<OrientationTypeNavigator*>(&type))
            {
                return visit(*t, key, dynamic_cast<NDArrayDataNavigator&>(data));
            }

            ARMARX_CHECK(false) << "Unhandled AronTypeNavigatorType '" << simox::meta::get_type_name(type) << "'."
                                << "\n(Data: '" << simox::meta::get_type_name(data) << "')";
        }
        catch (const std::bad_cast& e)
        {
            std::stringstream msg;
            msg << "Got ARON data '" << simox::meta::get_type_name(data) << "'"
                << " incompatible to ARON type '" << simox::meta::get_type_name(type) << "'"
                << " (got std::bad_cast: '" << e.what() << "').";
            throw error::AronException("TypedDataVisitor", __FUNCTION__, msg.str());
        }
    }

    bool TypedDataVisitor::applyToChildren(ObjectTypeNavigator& type, DictDataNavigator& data)
    {
        for (const std::string& key : type.getAllKeys())
        {
            TypeNavigator::PointerType childType = type.getMemberType(key);
            DataNavigator::PointerType childData = data.getElement(key);
            if (childType && childData)
            {
                if (!applyTo(key, *childType, *childData))
                {
                    return false;
                }
            }
        }
        return true;
    }

    bool TypedDataVisitor::applyToChildren(DictTypeNavigator& type, DictDataNavigator& data)
    {
        TypeNavigator::PointerType childType = type.getAcceptedType();
        for (const std::string& key : data.getAllKeys())
        {
            DataNavigator::PointerType childData = data.getElement(key);
            if (childType && childData)
            {
                if (!applyTo(key, *childType, *childData))
                {
                    return false;
                }
            }
        }
        return true;
    }

    bool TypedDataVisitor::applyToChildren(ListTypeNavigator& type, ListDataNavigator& data)
    {
        TypeNavigator::PointerType childType = type.getAcceptedType();
        size_t i = 0;
        for (const DataNavigator::PointerType& childData : data.getElements())
        {
            if (childType && childData)
            {
                if (!applyTo(std::to_string(i), *childType, *childData))
                {
                    return false;
                }
            }
            ++i;
        }
        return true;
    }

    bool TypedDataVisitor::applyToChildren(TupleTypeNavigator& type, ListDataNavigator& data)
    {
        unsigned int i = 0;
        for (const TypeNavigator::PointerType& childType : type.getAcceptedTypes())
        {
            DataNavigator::PointerType childData = data.getElement(i);
            if (childType && childData)
            {
                if (!applyTo(std::to_string(i), *childType, *childData))
                {
                    return false;
                }
            }
            ++i;
        }
        return true;
    }

}
